/*
  histogram.hpp

  Copyright (C) 2016 Taiji Yamada <taiji@aihara.co.jp>
*/
#ifndef _HISTOGRAM_HPP_
#define _HISTOGRAM_HPP_

#if defined(USE_LONG_DOUBLE_AS_INTERNAL_VALUE_TYPE)
#define INTERNAL_VALUE_TYPE long double
#else
#define INTERNAL_VALUE_TYPE T
#endif

template <typename InputIterator>
size_t histogram(std::vector<size_t> &heights,
		 const InputIterator first, const InputIterator last,
		 const typename std::iterator_traits<InputIterator>::value_type lb,
		 const typename std::iterator_traits<InputIterator>::value_type ub,
		 const bool logarithmic = false)
{
  using std::log;
  size_t rv(0);
  const typename std::iterator_traits<InputIterator>::value_type sz(!logarithmic ? ub - lb : log(ub) - log(lb));
  for (InputIterator it(first); it!=last; ++it) {
    if (*it < lb) continue;
    const typename std::iterator_traits<InputIterator>::value_type v(!logarithmic ? *it - lb : log(*it) - log(lb));
    const size_t id(v/sz*heights.size());
    if (!(id < heights.size())) continue;
    ++heights[id];
    ++rv;
  }
  return rv;
}
template <typename T>
inline
T histogram_lower_bound(const size_t i, const size_t n_bins, const T lb, const T ub, const bool logarithmic = false, const bool option_lt_ub = false)
{
  using std::exp;
  using std::log;
#if 0
  if (option_lt_ub)
    return !logarithmic ? lb + (ub - lb)/n_bins*i : lb*exp((log(ub) - log(lb))*i/n_bins);
  else
    return !logarithmic ? T(n_bins - i)/n_bins*lb + T(i)/n_bins*ub : exp((T)(n_bins - i)/n_bins*log(lb) + (T)i/n_bins*log(ub));
#else
  if (option_lt_ub)
    return !logarithmic ? lb + (ub - lb)/n_bins*i : lb*exp((log((INTERNAL_VALUE_TYPE)ub) - log((INTERNAL_VALUE_TYPE)lb))*i/n_bins);
  else
    return !logarithmic ? T(n_bins - i)/n_bins*lb + T(i)/n_bins*ub : exp((INTERNAL_VALUE_TYPE)(n_bins - i)/n_bins*log((INTERNAL_VALUE_TYPE)lb) + (INTERNAL_VALUE_TYPE)i/n_bins*log((INTERNAL_VALUE_TYPE)ub));
#endif
}
template <typename T>
inline
T histogram_upper_bound(const size_t i, const size_t n_bins, const T lb, const T ub, const bool logarithmic = false, const bool option_lt_ub = false)
{
  return histogram_lower_bound(i+1, n_bins, lb, ub, logarithmic, option_lt_ub);
}
template <typename T>
inline
std::pair<T, T> histogram_range(const size_t i, const size_t n_bins, const T lb, const T ub, const bool logarithmic = false, const bool option_lt_ub = false)
{
  return std::make_pair(histogram_lower_bound(i, n_bins, lb, ub, logarithmic, option_lt_ub),
			histogram_upper_bound(i, n_bins, lb, ub, logarithmic, option_lt_ub));
}
template <typename T>
T histogram_mean(const std::vector<size_t> &heights, const T lb, const T ub, const bool logarithmic = false)
{
  INTERNAL_VALUE_TYPE w(0), v(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    if (!(*it > 0)) continue;
    const std::pair<T, T> range(histogram_range(std::distance(heights.begin(), it), heights.size(), lb, ub, logarithmic));
#if 0
    const INTERNAL_VALUE_TYPE p((range.first + range.second)/2);
#else
    const INTERNAL_VALUE_TYPE p(!logarithmic ? (range.first + range.second)/2 : exp((log(range.first) + log(range.second))/2));
#endif
    w += *it;
    v += (p - v)*((INTERNAL_VALUE_TYPE)*it/w);
  }
  return v;
}
template <typename T>
T histogram_variance(const std::vector<size_t> &heights, const T lb, const T ub, const bool logarithmic = false)
{
  using std::exp;
  using std::log;
  INTERNAL_VALUE_TYPE mu(histogram_mean(heights, lb, ub, logarithmic)), w(0), v(0);
  for (std::vector<size_t>::const_iterator it=heights.begin(); it!=heights.end(); ++it) {
    if (!(*it > 0)) continue;
    const std::pair<T, T> range(histogram_range(std::distance(heights.begin(), it), heights.size(), lb, ub, logarithmic));
#if 0
    const INTERNAL_VALUE_TYPE p((range.first + range.second)/2), delta(p - mu);
#else
    const INTERNAL_VALUE_TYPE p(!logarithmic ? (range.first + range.second)/2 : exp((log(range.first) + log(range.second))/2)), delta(p - mu);
#endif
    w += *it;
    v += (delta*delta - v)*((INTERNAL_VALUE_TYPE)*it/w);
  }
  return v;
}
template <typename T>
T histogram_standard_deviation(const std::vector<size_t> &heights, const T lb, const T ub, const bool logarithmic = false)
{
  using std::sqrt;
  return sqrt(histogram_variance(heights, lb, ub, logarithmic));
}

#endif
